# This file is a modified copy of https://github.com/takluyver/bash_kernel/blob/0966dc102d7549f5c909c93de633a95b2af9f707/bash_kernel/install.py

# Copyright (c) 2015, Thomas Kluyver and contributors 
# All rights reserved.

# BSD 3-clause license:

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import os
import sys
import argparse

from jupyter_client.kernelspec import KernelSpecManager
from IPython.utils.tempdir import TemporaryDirectory
from powershell_kernel.util import get_powershell

kernel_json = {
 "argv": [sys.executable, "-m", "powershell_kernel", "-f", "{connection_file}"],
 "display_name": "PowerShell",
 "language": "powershell",
}

def install_my_kernel_spec(user=True, prefix=None, powershell_command=None):
    
    if sys.version_info >= (3, 0): 
        if powershell_command is None:
            powershell_command = get_powershell()
        kernel_json.update({'env': {'powershell_command' : powershell_command}})
        print('Using powershell_command=%r' % powershell_command)
    else:
        # python 2 cannot use env to pass values to the kernel
        # https://github.com/vors/jupyter-powershell/issues/7
        # TODO(python2): find a way to pass it
        if powershell_command is not None:
            print('Ignoring powershell_command on python2, jupyter will use default powershell_command=%r' % powershell_command)

    with TemporaryDirectory() as td:
        os.chmod(td, 0o755) # Starts off as 700, not user readable
        with open(os.path.join(td, 'kernel.json'), 'w') as f:
            json.dump(kernel_json, f, sort_keys=True)
        # TODO: Copy resources once they're specified

        print('Installing IPython kernel spec')
        KernelSpecManager().install_kernel_spec(td, 'powershell', user=user, prefix=prefix)

def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False # assume not an admin on non-Unix platforms

def main(argv=None):
    parser = argparse.ArgumentParser(
        description='Install KernelSpec for PowerShell Kernel'
    )
    prefix_locations = parser.add_mutually_exclusive_group()

    prefix_locations.add_argument(
        '--user',
        help='Install KernelSpec in user homedirectory',
        action='store_true'
    )
    prefix_locations.add_argument(
        '--sys-prefix',
        help='Install KernelSpec in sys.prefix. Useful in conda / virtualenv',
        action='store_true',
        dest='sys_prefix'
    )
    prefix_locations.add_argument(
        '--prefix',
        help='Install KernelSpec in this prefix',
        default=None
    )
    prefix_locations.add_argument(
        '--powershell-command',
        help='Command to run powershell (by default "powershell" on windows and "pwsh" on unix)',
        default=None
    )

    args = parser.parse_args(argv)

    user = False
    prefix = None
    powershell_command = None
    if args.sys_prefix:
        prefix = sys.prefix
    elif args.prefix:
        prefix = args.prefix
    elif args.user or not _is_root():
        user = True

    if args.powershell_command:
        powershell_command = args.powershell_command

    install_my_kernel_spec(user=user, prefix=prefix, powershell_command=powershell_command)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright (c) 2011, Wojciech Bederski (wuub.net)
# All rights reserved.
# See LICENSE.txt for details.
from __future__ import absolute_import, unicode_literals, print_function, division

import subprocess
import os
import sys
import re
import signal
from subprocess import Popen
from codecs import getencoder, getincrementaldecoder

PY3 = sys.version_info[0] == 3

if os.name == 'posix':
    POSIX = True
    import fcntl
    import select
else:
    POSIX = False

class SubprocessRepl(object):
    def __init__(self, cmd):
        self.encoder = getencoder('utf8')
        self.decoder = getincrementaldecoder('utf8')()
        self.popen = Popen(cmd, bufsize=1,
            stderr=subprocess.STDOUT, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        if POSIX:
            flags = fcntl.fcntl(self.popen.stdout, fcntl.F_GETFL)
            fcntl.fcntl(self.popen.stdout, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    def is_alive(self):
        return self.popen.poll() is None

    def read_bytes(self):
        out = self.popen.stdout
        if POSIX:
            while True:
                i, _, _ = select.select([out], [], [])
                if i:
                    return out.read(4096)
        else:
            # this is windows specific problem, that you cannot tell if there
            # are more bytes ready, so we read only 1 at a times

            while True:
                byte = self.popen.stdout.read(1)
                if byte == b'\r':
                    # f'in HACK, for \r\n -> \n translation on windows
                    # I tried universal_endlines but it was pain and misery! :'(
                    continue
                return byte

    def write(self, command):
        (bytes, how_many) = self.encoder(command)
        si = self.popen.stdin
        si.write(bytes)
        si.flush()
    
    def reset_decoder(self):
        self.decoder = getincrementaldecoder('utf8')()

    def read(self):
        """Reads at least one decoded char of output"""
        while True:
            bs = self.read_bytes()
            if not bs:
                return None
            try:
                output = self.decoder.decode(bs)
            except Exception as e:
                output = "■"
                self.reset_decoder()
            if output:
                return output
  


from ipykernel.kernelbase import Kernel

from os import unlink, environ

import base64
import imghdr
import re
import signal
import urllib
import sys
import traceback

from powershell_kernel import subprocess_repl, powershell_proxy
from powershell_kernel.util import get_powershell

__version__ = '0.1.4'

version_pat = re.compile(r'version (\d+(\.\d+)+)')

class PowerShellKernel(Kernel):
    implementation = 'powershell_kernel'
    implementation_version = __version__

    @property
    def language_version(self):
        m = version_pat.search(self.banner)
        return m.group(1)

    _banner = None

    @property
    def banner(self):
        return self._banner

    language_info = {'name': 'powershell',
                     'codemirror_mode': 'shell',
                     'mimetype': 'text/x-sh',
                     'file_extension': '.ps1'}

    proxy = None

    def __createProxy(self):
        # powershell_command env variable is set by the kernel to allow both powershell and pwsh
        # but on python2 we cannot pass it thru env variable, see https://github.com/vors/jupyter-powershell/issues/7
        # TODO(python2): can we pass it somehow differently and still provide user-picked value on python2?
        try:
            powershell_command = environ['powershell_command']
        except:
            powershell_command = get_powershell()

        repl = subprocess_repl.SubprocessRepl([powershell_command, '-noninteractive', '-noprofile', '-File', '-'])
        self.proxy = powershell_proxy.ReplProxy(repl)

    def do_execute(self, code, silent, store_history=True,
                   user_expressions=None, allow_stdin=False):
        try:
            if not code.strip():
                return {'status': 'ok', 'execution_count': self.execution_count,
                        'payload': [], 'user_expressions': {}}

            if not self.proxy:
                self.__createProxy()

            for output in self.proxy.run_command(code):
                message = {'name': 'stdout', 'text': output}
                self.send_response(self.iopub_socket, 'stream', message)

            return {'status': 'ok', 'execution_count': self.execution_count,
                    'payload': [], 'user_expressions': {}}

        except Exception:
            excInfo = sys.exc_info()
            message = {
                'ename': str(excInfo[0].__name__),
                'evalue': str(excInfo[1]),
                'traceback': [traceback.format_exc()]
            }
            self.send_response(self.iopub_socket, 'error', message)
            message['status'] = 'error'
            return message



import threading
try:
    import queue
except ImportError:
    import Queue as queue
from threading import Timer, Lock
from time import sleep

class ReplReader(threading.Thread):
    def __init__(self, repl):
        super(ReplReader, self).__init__()
        self.repl = repl
        self.daemon = True
        self.queue = queue.Queue()
        self.start()

    def run(self):
        r = self.repl
        q = self.queue
        while True:
            result = r.read()
            q.put(result)
            if result is None:
                break

class ReplProxy(object):
    def __init__(self, repl):
        self.runCmdLock = Lock()

        self._repl = repl
        self._repl_reader = ReplReader(repl)

        self.stop_flag = False
        self.output = ''
        self.timer = Timer(0.1, self.update_view_loop)
        self.timer.start()

        self.output_prefix_stripped = True
        self.expected_output_prefix = ''
        self.expected_output_len = 0

        # Returns a generator that yields string messages as they are returned from powershell via stdout
        # this is a hack to detect when we stop processing this input
        for temp in self.run_command('function prompt() {"^"}'):
            continue
        
    def run_command(self, input):
        self.runCmdLock.acquire()
        try:
            self.output = ''
            self.stop_flag = False

            # Append newline to the original input to handle single line comments on the last line
            #
            # Also, for multiline statements we should send 1 extra new line at the end
            # https://stackoverflow.com/questions/13229066/how-to-end-a-multi-line-command-in-powershell
            input = '. {\n' + input + '\n}\n'

            self.expected_output_prefix = input.replace('\n', '\n>> ') + '\n'
            self.expected_output_len = len(self.expected_output_prefix)
            self.output_prefix_stripped = False

            self._repl.write(input + '\n')
            while not self.stop_flag:
                sleep(0.05)
                # Allows for interactive streaming of output
                if not self.stop_flag:
                    powershell_message = self.output
                    self.output = ''
                    if powershell_message != '':
                        yield powershell_message
            yield self.output

        finally:
            self.runCmdLock.release()

    def handle_repl_output(self):
        """Returns new data from Repl and bool indicating if Repl is still
           working"""
        if self.stop_flag:
            return True
        try:
            while True:
                packet = self._repl_reader.queue.get_nowait()
                if packet is None:
                    return False

                self.write(packet)

        except queue.Empty:
            return True

    def update_view_loop(self):
        is_still_working = self.handle_repl_output()
        if is_still_working:
            self.timer = Timer(0.1, self.update_view_loop)
            self.timer.start()
        else:
            self.write("\n***Repl Killed***\n""")

    def write(self, packet):
        # this is a hack to detect when we stop processing this input
        if packet.endswith('^'):
            self.stop_flag = True
            self.output += packet[:-1]
            return
        self.output += packet

        if not self.output_prefix_stripped and len(self.output) >= self.expected_output_len:
            if self.output[:self.expected_output_len] != self.expected_output_prefix:
                print("Unexpected prefix: %r : Expected %r" % (
                    self.output[:self.expected_output_len], self.expected_output_prefix
                ))
            else:
                self.output_prefix_stripped = True
                self.output = self.output[self.expected_output_len:]



#from distutils.core import setup
from setuptools import setup
setup(
  name = 'powershell_kernel',
  packages = ['powershell_kernel'],
  version = '0.1.4',
  description = 'PowerShell language kernel for Jupyter',
  author = 'Sergei Vorobev',
  author_email = 'xvorsx@gmail.com',
  url = 'https://github.com/vors/jupyter-powershell',
  download_url = 'https://github.com/vors/jupyter-powershell/archive/0.1.4.tar.gz',
  keywords = ['ikernel', 'jupyter', 'powershell', 'pwsh'],
  classifiers = [],
)
